@@ -26,18 +26,27 @@ try:
 except ImportError:
     numpy = None
 
+HAS_NUMPY_FINANCIAL = False
+try:
+    import numpy_financial
+    HAS_NUMPY_FINANCIAL = True
+except ImportError:
+    pass
+
+
+
 MAX_EXPONENT = 10000
 MAX_STR_LEN = 2 << 17  # 256KiB
 MAX_SHIFT = 1000
 MAX_OPEN_BUFFER = 2 << 17
 
-RESERVED_WORDS = ('and', 'as', 'assert', 'break', 'class', 'continue',
-                  'def', 'del', 'elif', 'else', 'except', 'exec',
-                  'finally', 'for', 'from', 'global', 'if', 'import',
-                  'in', 'is', 'lambda', 'not', 'or', 'pass', 'print',
-                  'raise', 'return', 'try', 'while', 'with', 'True',
-                  'False', 'None', 'eval', 'execfile', '__import__',
-                  '__package__')
+RESERVED_WORDS = ('False', 'None', 'True', 'and', 'as', 'assert',
+                  'async', 'await', 'break', 'class', 'continue',
+                  'def', 'del', 'elif', 'else', 'except', 'finally',
+                  'for', 'from', 'global', 'if', 'import', 'in', 'is',
+                  'lambda', 'nonlocal', 'not', 'or', 'pass', 'raise',
+                  'return', 'try', 'while', 'with', 'yield', 'exec',
+                  'eval', 'execfile', '__import__', '__package__')
 
 NAME_MATCH = re.compile(r"[a-zA-Z_][a-zA-Z0-9_]*$").match
 
@@ -85,82 +94,78 @@ FROM_MATH = ('acos', 'acosh', 'asin', 'asinh', 'atan', 'atan2', 'atanh',
 MATH_TABLE = {sym: getattr(math, sym) for sym in FROM_MATH if hasattr(math, sym)}
 
 
-FROM_NUMPY = ('Inf', 'NAN', 'abs', 'add', 'alen', 'all', 'amax', 'amin',
-              'angle', 'any', 'append', 'arange', 'arccos', 'arccosh',
-              'arcsin', 'arcsinh', 'arctan', 'arctan2', 'arctanh',
-              'argmax', 'argmin', 'argsort', 'argwhere', 'around', 'array',
-              'array2string', 'asanyarray', 'asarray', 'asarray_chkfinite',
-              'ascontiguousarray', 'asfarray', 'asfortranarray',
-              'asmatrix', 'asscalar', 'atleast_1d', 'atleast_2d',
-              'atleast_3d', 'average', 'bartlett', 'base_repr',
-              'bitwise_and', 'bitwise_not', 'bitwise_or', 'bitwise_xor',
-              'blackman', 'bool', 'broadcast', 'broadcast_arrays', 'byte',
-              'c_', 'cdouble', 'ceil', 'cfloat', 'chararray', 'choose',
+FROM_NUMPY = ('Inf', 'NAN', 'abs', 'add', 'all', 'amax', 'amin', 'angle',
+              'any', 'append', 'arange', 'arccos', 'arccosh', 'arcsin',
+              'arcsinh', 'arctan', 'arctan2', 'arctanh', 'argmax', 'argmin',
+              'argsort', 'argwhere', 'around', 'array', 'array2string',
+              'asanyarray', 'asarray', 'asarray_chkfinite',
+              'ascontiguousarray', 'asfarray', 'asfortranarray', 'asmatrix',
+              'atleast_1d', 'atleast_2d', 'atleast_3d', 'average', 'bartlett',
+              'base_repr', 'bitwise_and', 'bitwise_not', 'bitwise_or',
+              'bitwise_xor', 'blackman', 'broadcast', 'broadcast_arrays',
+              'byte', 'c_', 'cdouble', 'ceil', 'cfloat', 'chararray', 'choose',
               'clip', 'clongdouble', 'clongfloat', 'column_stack',
-              'common_type', 'complex', 'complex128', 'complex64',
-              'complex_', 'complexfloating', 'compress', 'concatenate',
-              'conjugate', 'convolve', 'copy', 'copysign', 'corrcoef',
-              'correlate', 'cos', 'cosh', 'cov', 'cross', 'csingle',
-              'cumprod', 'cumsum', 'datetime_data', 'deg2rad', 'degrees',
-              'delete', 'diag', 'diag_indices', 'diag_indices_from',
-              'diagflat', 'diagonal', 'diff', 'digitize', 'divide', 'dot',
-              'double', 'dsplit', 'dstack', 'dtype', 'e', 'ediff1d',
-              'empty', 'empty_like', 'equal', 'exp', 'exp2', 'expand_dims',
-              'expm1', 'extract', 'eye', 'fabs', 'fill_diagonal', 'finfo',
-              'fix', 'flatiter', 'flatnonzero', 'fliplr', 'flipud',
-              'float', 'float32', 'float64', 'float_', 'floating', 'floor',
-              'floor_divide', 'fmax', 'fmin', 'fmod', 'format_parser',
-              'frexp', 'frombuffer', 'fromfile', 'fromfunction',
-              'fromiter', 'frompyfunc', 'fromregex', 'fromstring', 'fv',
-              'genfromtxt', 'getbufsize', 'geterr', 'gradient', 'greater',
-              'greater_equal', 'hamming', 'hanning', 'histogram',
-              'histogram2d', 'histogramdd', 'hsplit', 'hstack', 'hypot',
-              'i0', 'identity', 'iinfo', 'imag', 'in1d', 'index_exp',
-              'indices', 'inexact', 'inf', 'info', 'infty', 'inner',
-              'insert', 'int', 'int0', 'int16', 'int32', 'int64', 'int8',
-              'int_', 'int_asbuffer', 'intc', 'integer', 'interp',
-              'intersect1d', 'intp', 'invert', 'ipmt', 'irr', 'iscomplex',
-              'iscomplexobj', 'isfinite', 'isfortran', 'isinf', 'isnan',
-              'isneginf', 'isposinf', 'isreal', 'isrealobj', 'isscalar',
-              'issctype', 'iterable', 'ix_', 'kaiser', 'kron', 'ldexp',
-              'left_shift', 'less', 'less_equal', 'linspace',
-              'little_endian', 'load', 'loads', 'loadtxt', 'log', 'log10',
-              'log1p', 'log2', 'logaddexp', 'logaddexp2', 'logical_and',
-              'logical_not', 'logical_or', 'logical_xor', 'logspace',
-              'long', 'longcomplex', 'longdouble', 'longfloat', 'longlong',
-              'mafromtxt', 'mask_indices', 'mat', 'matrix',
+              'common_type', 'complex128', 'complex64', 'complex_',
+              'complexfloating', 'compress', 'concatenate', 'conjugate',
+              'convolve', 'copy', 'copysign', 'corrcoef', 'correlate', 'cos',
+              'cosh', 'cov', 'cross', 'csingle', 'cumprod', 'cumsum',
+              'datetime_data', 'deg2rad', 'degrees', 'delete', 'diag',
+              'diag_indices', 'diag_indices_from', 'diagflat', 'diagonal',
+              'diff', 'digitize', 'divide', 'dot', 'double', 'dsplit',
+              'dstack', 'dtype', 'e', 'ediff1d', 'empty', 'empty_like',
+              'equal', 'exp', 'exp2', 'expand_dims', 'expm1', 'extract', 'eye',
+              'fabs', 'fft', 'fill_diagonal', 'finfo', 'fix', 'flatiter',
+              'flatnonzero', 'fliplr', 'flipud', 'float32', 'float64',
+              'float_', 'floating', 'floor', 'floor_divide', 'fmax', 'fmin',
+              'fmod', 'format_parser', 'frexp', 'frombuffer', 'fromfile',
+              'fromfunction', 'fromiter', 'frompyfunc', 'fromregex',
+              'fromstring', 'genfromtxt', 'getbufsize', 'geterr', 'gradient',
+              'greater', 'greater_equal', 'hamming', 'hanning', 'histogram',
+              'histogram2d', 'histogramdd', 'hsplit', 'hstack', 'hypot', 'i0',
+              'identity', 'iinfo', 'imag', 'in1d', 'index_exp', 'indices',
+              'inexact', 'inf', 'info', 'infty', 'inner', 'insert', 'int16',
+              'int32', 'int64', 'int8', 'int_', 'intc', 'integer', 'interp',
+              'intersect1d', 'intp', 'invert', 'iscomplex', 'iscomplexobj',
+              'isfinite', 'isfortran', 'isinf', 'isnan', 'isneginf',
+              'isposinf', 'isreal', 'isrealobj', 'isscalar', 'issctype',
+              'iterable', 'ix_', 'kaiser', 'kron', 'ldexp', 'left_shift',
+              'less', 'less_equal', 'linalg', 'linspace', 'little_endian',
+              'load', 'loadtxt', 'log', 'log10', 'log1p', 'log2', 'logaddexp',
+              'logaddexp2', 'logical_and', 'logical_not', 'logical_or',
+              'logical_xor', 'logspace', 'longcomplex', 'longdouble',
+              'longfloat', 'longlong', 'mask_indices', 'mat', 'matrix',
               'maximum', 'maximum_sctype', 'may_share_memory', 'mean',
               'median', 'memmap', 'meshgrid', 'mgrid', 'minimum',
-              'mintypecode', 'mirr', 'mod', 'modf', 'msort', 'multiply',
-              'nan', 'nan_to_num', 'nanargmax', 'nanargmin', 'nanmax',
-              'nanmin', 'nansum', 'ndarray', 'ndenumerate', 'ndfromtxt',
-              'ndim', 'ndindex', 'negative', 'newaxis', 'nextafter',
-              'nonzero', 'not_equal', 'nper', 'npv', 'number',
-              'obj2sctype', 'ogrid', 'ones', 'ones_like', 'outer',
-              'packbits', 'percentile', 'pi', 'piecewise', 'place', 'pmt',
-              'poly', 'poly1d', 'polyadd', 'polyder', 'polydiv', 'polyfit',
-              'polyint', 'polymul', 'polysub', 'polyval', 'power', 'ppmt',
-              'prod', 'product', 'ptp', 'put', 'putmask', 'pv', 'r_',
-              'rad2deg', 'radians', 'rank', 'rate', 'ravel', 'real',
-              'real_if_close', 'reciprocal', 'record', 'remainder',
-              'repeat', 'reshape', 'resize', 'restoredot', 'right_shift',
-              'rint', 'roll', 'rollaxis', 'roots', 'rot90', 'round',
-              'round_', 'row_stack', 's_', 'sctype2char', 'searchsorted',
-              'select', 'setbufsize', 'setdiff1d', 'seterr', 'setxor1d',
-              'shape', 'short', 'sign', 'signbit', 'signedinteger', 'sin',
-              'sinc', 'single', 'singlecomplex', 'sinh', 'size',
-              'sometrue', 'sort', 'sort_complex', 'spacing', 'split',
-              'sqrt', 'square', 'squeeze', 'std', 'str', 'str_',
-              'subtract', 'sum', 'swapaxes', 'take', 'tan', 'tanh',
-              'tensordot', 'tile', 'trace', 'transpose', 'trapz', 'tri',
-              'tril', 'tril_indices', 'tril_indices_from', 'trim_zeros',
-              'triu', 'triu_indices', 'triu_indices_from', 'true_divide',
-              'trunc', 'ubyte', 'uint', 'uint0', 'uint16', 'uint32',
-              'uint64', 'uint8', 'uintc', 'uintp', 'ulonglong', 'union1d',
-              'unique', 'unravel_index', 'unsignedinteger', 'unwrap',
-              'ushort', 'vander', 'var', 'vdot', 'vectorize', 'vsplit',
-              'vstack', 'where', 'who', 'zeros', 'zeros_like',
-              'fft', 'linalg', 'polynomial', 'random')
+              'mintypecode', 'mod', 'modf', 'msort', 'multiply', 'nan',
+              'nan_to_num', 'nanargmax', 'nanargmin', 'nanmax', 'nanmin',
+              'nansum', 'ndarray', 'ndenumerate', 'ndim', 'ndindex',
+              'negative', 'newaxis', 'nextafter', 'nonzero', 'not_equal',
+              'number', 'obj2sctype', 'ogrid', 'ones', 'ones_like', 'outer',
+              'packbits', 'percentile', 'pi', 'piecewise', 'place', 'poly',
+              'poly1d', 'polyadd', 'polyder', 'polydiv', 'polyfit', 'polyint',
+              'polymul', 'polynomial', 'polysub', 'polyval', 'power', 'prod',
+              'product', 'ptp', 'put', 'putmask', 'r_', 'rad2deg', 'radians',
+              'random', 'ravel', 'real', 'real_if_close', 'reciprocal',
+              'record', 'remainder', 'repeat', 'reshape', 'resize',
+              'right_shift', 'rint', 'roll', 'rollaxis', 'roots', 'rot90',
+              'round', 'round_', 'row_stack', 's_', 'sctype2char',
+              'searchsorted', 'select', 'setbufsize', 'setdiff1d', 'seterr',
+              'setxor1d', 'shape', 'short', 'sign', 'signbit', 'signedinteger',
+              'sin', 'sinc', 'single', 'singlecomplex', 'sinh', 'size',
+              'sometrue', 'sort', 'sort_complex', 'spacing', 'split', 'sqrt',
+              'square', 'squeeze', 'std', 'str_', 'subtract', 'sum',
+              'swapaxes', 'take', 'tan', 'tanh', 'tensordot', 'tile', 'trace',
+              'transpose', 'trapz', 'tri', 'tril', 'tril_indices',
+              'tril_indices_from', 'trim_zeros', 'triu', 'triu_indices',
+              'triu_indices_from', 'true_divide', 'trunc', 'ubyte', 'uint',
+              'uint16', 'uint32', 'uint64', 'uint8', 'uintc', 'uintp',
+              'ulonglong', 'union1d', 'unique', 'unravel_index',
+              'unsignedinteger', 'unwrap', 'ushort', 'vander', 'var', 'vdot',
+              'vectorize', 'vsplit', 'vstack', 'where', 'who', 'zeros',
+              'zeros_like')
+
+FROM_NUMPY_FINANCIAL = ('fv', 'ipmt', 'irr', 'mirr', 'nper', 'npv',
+                        'pmt', 'ppmt', 'pv', 'rate')
 
 NUMPY_RENAMES = {'ln': 'log', 'asin': 'arcsin', 'acos': 'arccos',
                  'atan': 'arctan', 'atan2': 'arctan2', 'atanh':
@@ -187,6 +192,11 @@ if HAS_NUMPY:
         NUMPY_TABLE[sym] = getattr(numpy, sym)
     for sname, sym in NUMPY_RENAMES.items():
         NUMPY_TABLE[sname] = getattr(numpy, sym)
+
+    if HAS_NUMPY_FINANCIAL:
+        for sym in FROM_NUMPY_FINANCIAL:
+            NUMPY_TABLE[sym] = getattr(numpy_financial, sym)
+
 else:
     NUMPY_TABLE = {}
 
@@ -312,19 +322,20 @@ class Empty:
     """Empty class."""
     def __init__(self):
         """TODO: docstring in public method."""
-        pass
+        return
 
     def __nonzero__(self):
         """Empty is TODO: docstring in magic method."""
         return False
 
+    def __repr__(self):
+        """Empty is TODO: docstring in magic method."""
+        return "Empty"
 
 ReturnedNone = Empty()
 
-
 class ExceptionHolder:
     """Basic exception handler."""
-
     def __init__(self, node, exc=None, msg='', expr=None, lineno=None):
         """TODO: docstring in public method."""
         self.node = node
@@ -371,7 +382,7 @@ class NameFinder(ast.NodeVisitor):
     def generic_visit(self, node):
         """TODO: docstring in public method."""
         if node.__class__.__name__ == 'Name':
-            if node.ctx.__class__ == ast.Load and node.id not in self.names:
+            if node.id not in self.names:
                 self.names.append(node.id)
         ast.NodeVisitor.generic_visit(self, node)
 
@@ -383,35 +394,119 @@ def get_ast_names(astnode):
     return finder.names
 
 
-def make_symbol_table(use_numpy=True, **kws):
+def valid_varname(name):
+    "is this a valid variable name"
+    return name.isidentifier() and name not in RESERVED_WORDS
+
+
+class Group(dict):
+    """
+    Group: a container of objects that can be accessed either as an object attributes
+    or dictionary  key/value.  Attribute names must follow Python naming conventions.
+
+
+    """
+    def __init__(self, name=None, searchgroups=None, **kws):
+        if name is None:
+            name = hex(id(self))
+        self.__name__ = name
+        dict.__init__(self, **kws)
+        self._searchgroups = searchgroups
+
+    def __setattr__(self, name, value):
+        if not valid_varname(name):
+            raise SyntaxError(f"invalid attribute name '{name}'")
+        self[name] = value
+
+    def __getattr__(self, name, default=None):
+        if name in self:
+            return self[name]
+        if default is not None:
+            return default
+        raise KeyError(f"no attribute named '{name}'")
+
+    def __setitem__(self, name, value):
+        if not valid_varname(name):
+            raise SyntaxError(f"invalid attribute name '{name}'")
+        dict.__setitem__(self, name, value)
+
+    def get(self, key, default=None):
+        val = self.__getattr__(key, ReturnedNone)
+        if not isinstance(val, Empty):
+            return val
+        searchgroups = self._searchgroups
+        if searchgroups is not None:
+            for sgroup in searchgroups:
+                grp = self.__getattr__(sgroup, None)
+                if isinstance(grp, (Group, dict)):
+                    val = grp.__getattr__(key, ReturnedNone)
+                    if not isinstance(val, Empty):
+                        return val
+        return default
+
+
+    def __repr__(self):
+        keys = [a for a in self.keys() if a != '__name__']
+        return f"Group('{self.__name__}', {len(keys)} symbols)"
+
+    def _repr_html_(self):
+        """HTML representation for Jupyter notebook"""
+        html = [f"<table><caption>Group('{self.__name__}')</caption>",
+  "<tr><th>Attribute</th><th>DataType</th><th><b>Value</b></th></tr>"]
+        for key, val in self.items():
+            html.append(f"""
+<tr><td>{key}</td><td><i>{type(val).__name__}</i></td>
+    <td>{repr(val):.75s}</td>
+</tr>""")
+        html.append("</table>")
+        return '\n'.join(html)
+
+
+def make_symbol_table(use_numpy=True, nested=False, top=True,  **kws):
     """Create a default symboltable, taking dict of user-defined symbols.
 
     Arguments
     ---------
     numpy : bool, optional
-       whether to include symbols from numpy
+       whether to include symbols from numpy [True]
+    nested : bool, optional
+       whether to make a "new-style" nested table instead of a plain dict [False]
+    top : bool, optional
+       whether this is the top-level table in a nested-table [True]
     kws :  optional
        additional symbol name, value pairs to include in symbol table
 
     Returns
     --------
-    symbol_table : dict
+    symbol_table : dict or nested Group
        a symbol table that can be used in `asteval.Interpereter`
 
     """
-    symtable = {}
+    if nested:
+        if top:
+            name = '_main'
+            if 'name' in kws:
+                name = kws.pop('name')
+        symtable = Group(name=name)
+    else:
+        symtable = {}
 
     symtable.update(BUILTINS_TABLE)
-    symtable.update(MATH_TABLE)
-    if use_numpy:
-        symtable.update(NUMPY_TABLE)
     symtable.update(LOCALFUNCS)
     symtable.update(kws)
+    math_functions = dict(MATH_TABLE.items())
+    if use_numpy:
+        math_functions.update(NUMPY_TABLE)
 
+    if nested:
+        symtable['math'] = Group(name='math', **math_functions)
+        symtable._searchgroups = ('math',)
+    else:
+        symtable.update(math_functions)
+    symtable.update(**kws)
     return symtable
 
 
-
 class Procedure:
     """Procedure: user-defined function for asteval.
 
@@ -469,11 +564,25 @@ class Procedure:
 
     def __call__(self, *args, **kwargs):
         """TODO: docstring in public method."""
-        symlocals = {}
+        topsym = self.__asteval__.symtable
+        if self.__asteval__.config.get('nested_symtable', False):
+            sargs = {'_main': topsym}
+            sgroups = topsym.get('_searchgroups', None)
+            if sgroups is not None:
+                for sxname in sgroups:
+                    sargs[sxname] = topsym.get(sxname)
+
+
+            symlocals = Group(name=f'symtable_{self.name}_', **sargs)
+            symlocals._searchgroups = list(sargs.keys())
+        else:
+            symlocals = {}
+
         args = list(args)
         nargs = len(args)
         nkws = len(kwargs)
         nargs_expected = len(self.argnames)
+
         # check for too few arguments, but the correct keyword given
         if (nargs < nargs_expected) and nkws > 0:
             for name in self.argnames[nargs:]:
@@ -539,8 +648,13 @@ class Procedure:
             msg = f"incorrect arguments for Procedure {self.name}"
             self.raise_exc(None, msg=msg, lineno=self.lineno)
 
-        save_symtable = self.__asteval__.symtable.copy()
-        self.__asteval__.symtable.update(symlocals)
+        if self.__asteval__.config.get('nested_symtable', False):
+            save_symtable = self.__asteval__.symtable
+            self.__asteval__.symtable = symlocals
+        else:
+            save_symtable = self.__asteval__.symtable.copy()
+            self.__asteval__.symtable.update(symlocals)
+
         self.__asteval__.retval = None
         self.__asteval__._calldepth += 1
         retval = None
